/* Copyright (c) 1999 Malmrose Software */
/* Version 1.0, May 25, 1999            */

#include <Application.h>
#include <InterfaceKit.h>
#include <StorageKit.h>
#include <TrackerAddOn.h>
#include <fs_attr.h>

// Uncomment one of the following lines to compile Cut or Copy or Paste
#define APP_CUT
//#define APP_COPY
//#define APP_PASTE

const char COPYRIGHT_STRING[] = "Copyright (c) 1999 Malmrose Software";
const char APP_SIG[] = "application/x-malmrosesoftware-xcv";
const char APP_NAME[] = "XCV";
const char MSG_CUT_NAME[] = "xcv-cutRefs";
const char MSG_COPY_NAME[] = "xcv-copyRefs";

BWindow *statusWindow = NULL;
BStatusBar *statusBar = NULL;

void MessageBox(const char *message);
void process_refs(entry_ref dir_ref, BMessage* msg, void*);

#if defined(APP_PASTE)
void HandleRefs(entry_ref &dir_ref, BMessage *sourceMsg, bool move);
void CopyFile(BEntry &sourceEntry, BDirectory &destDir);
void MoveFile(BEntry &sourceEntry, BDirectory &destDir);
void CopyDirectory(BEntry &sourceEntry, BDirectory &destDir);
void MoveDirectory(BEntry &sourceEntry, BDirectory &destDir);
void CopyLink(BEntry &sourceEntry, BDirectory &destDir);
void MoveLink(BEntry &sourceEntry, BDirectory &destDir);
void DoEntry(BEntry &entry, BDirectory destDir, bool move);
void DeleteDirectory(BDirectory &sourceDir);
int CountFile(BEntry &entry);
int CountDirectory(BEntry &entry);
int CountEntry(BEntry &entry);

const int BUFFER_SIZE = 65536;

void CopyAttributes(BNode &sourceFile, BNode &destFile)
	{
	char attrName[B_ATTR_NAME_LENGTH];
	attr_info ai;
	uint8 *atBuf = NULL;

	while (sourceFile.GetNextAttrName(attrName) == B_NO_ERROR)
		{
		sourceFile.GetAttrInfo(attrName, &ai);
		atBuf = new uint8[ai.size];
		sourceFile.ReadAttr(attrName, 0, 0, atBuf, ai.size);
		destFile.WriteAttr(attrName, ai.type, 0, atBuf, ai.size);
		delete[] atBuf;
		}
	}

void CopyStat(BStatable &sourceStatable, BStatable &destStatable)
	{
	status_t status;
	time_t ctime;

	status = sourceStatable.GetCreationTime(&ctime);
	destStatable.SetCreationTime(ctime);
	status = sourceStatable.GetModificationTime(&ctime);
	destStatable.SetModificationTime(ctime);
	status = sourceStatable.GetAccessTime(&ctime);
	destStatable.SetAccessTime(ctime);
	}

void CopyLink(BEntry &sourceEntry, BDirectory &destDir)
	{
	char name[B_FILE_NAME_LENGTH];
	char path[B_FILE_NAME_LENGTH];
	BSymLink sourceLink(&sourceEntry);
	BSymLink destLink;

	sourceEntry.GetName(name);
	sourceLink.ReadLink(path, B_FILE_NAME_LENGTH);
	destDir.CreateSymLink(name, path, &destLink);

	// Copy stat properties
	CopyStat(sourceEntry, destLink);
	CopyAttributes(sourceLink, destLink);

	statusWindow->Lock();
	statusBar->Update(1);
	statusWindow->Unlock();
	}

void MoveLink(BEntry &sourceEntry, BDirectory &destDir)
	{
	CopyLink(sourceEntry, destDir);
	sourceEntry.Remove();
	}

void CopyFile(BEntry &sourceEntry, BDirectory &destDir)
	{
	BFile destFile;
	BFile sourceFile(&sourceEntry, B_READ_ONLY);
	uint8 *buffer = NULL;
	ssize_t len;
	char name[B_FILE_NAME_LENGTH];

	sourceEntry.GetName(name);
	destDir.CreateFile(name, &destFile);
	buffer = new uint8[BUFFER_SIZE];

	// Copy data
	do
		{
		len = sourceFile.Read(buffer, BUFFER_SIZE);
		if (len)
			{
			destFile.Write(buffer, len);
			statusWindow->Lock();
			statusBar->Update(1);
			statusWindow->Unlock();
			}
		}
	while (len);

	// Set stat properties
	CopyStat(sourceEntry, destFile);
	CopyAttributes(sourceFile, destFile);

	delete[] buffer;
	}

void MoveFile(BEntry &sourceEntry, BDirectory &destDir)
	{
	BFile destFile;
	status_t status;

	status = sourceEntry.MoveTo(&destDir);
	switch (status)
		{
		case B_NO_ERROR:
			// Success.
			statusWindow->Lock();
			statusBar->Update(CountEntry(sourceEntry));
			statusWindow->Unlock();
			break;
		default:
			// Error. Try moving the file manually.
			// Copy the file.
			CopyFile(sourceEntry, destDir);
			
			// Delete the source.
			status = sourceEntry.Remove();
			if (status != B_NO_ERROR)
				MessageBox("Error removing source file.");
			break;
		}
	}

void CopyDirectory(BEntry &sourceEntry, BDirectory &destDir)
	{
	char name[B_FILE_NAME_LENGTH];
	BDirectory sourceDir;
	status_t status;
	BEntry tmpEntry;
	BDirectory endDir;

	// Create the destination directory.
	sourceEntry.GetName(name);
	destDir.CreateDirectory(name, &endDir);

	// Copy all the files and directories.
	status = sourceDir.SetTo(&sourceEntry);
	if (status == B_NO_ERROR)
		{
		// Success.
		sourceDir.Rewind();
		status = sourceDir.GetNextEntry(&tmpEntry);
		while (status != B_ENTRY_NOT_FOUND)
			{
			DoEntry(tmpEntry, endDir, false);
			status = sourceDir.GetNextEntry(&tmpEntry);
			}
		}

	// Set stat properties
	CopyStat(sourceEntry, endDir);
	CopyAttributes(sourceDir, endDir);

	statusWindow->Lock();
	statusBar->Update(1);
	statusWindow->Unlock();
	}

void DeleteDirectory(BDirectory &sourceDir)
	{
	BEntry tmpEntry;
	status_t status;
	BDirectory testDir;

	sourceDir.Rewind();
	status = sourceDir.GetNextEntry(&tmpEntry);
	while (status != B_ENTRY_NOT_FOUND)
		{
		if (testDir.SetTo(&tmpEntry) == B_OK)
			{
			DeleteDirectory(testDir);
			}
		else
			{
			status = tmpEntry.Remove();
			}
		status = sourceDir.GetNextEntry(&tmpEntry);
		}
	tmpEntry.SetTo(&sourceDir, ".");
	status = tmpEntry.Remove();
	}

void MoveDirectory(BEntry &sourceEntry, BDirectory &destDir)
	{
	BFile destFile;
	status_t status;
	BDirectory sourceDir;
	BDirectory testDir;
	BFile testFile;

	status = sourceEntry.MoveTo(&destDir);
	if (status != B_NO_ERROR)
		{
		// Error. Try moving the directory manually.
		// Copy the directory.
		CopyDirectory(sourceEntry, destDir);
		
		// Delete the source directory.
		status = sourceEntry.Remove();
		if (status != B_NO_ERROR)
			{
			// Error
			// Probably: directory contains files.

			// Delete all the files and directories.
			status = sourceDir.SetTo(&sourceEntry);
			if (status == B_NO_ERROR)
				{
				// Success.
				DeleteDirectory(sourceDir);
				}
			}
		}
	statusWindow->Lock();
	statusBar->Update(1);
	statusWindow->Unlock();
	}

void DoEntry(BEntry &entry, BDirectory destDir, bool move)
	{
	if (entry.IsDirectory())
		{
		// Copy/Move directory.
		if (move)
			MoveDirectory(entry, destDir);
		else
			CopyDirectory(entry, destDir);
		}
	else if (entry.IsFile())
		{
		// Copy/Move file.
		if (move)
			MoveFile(entry, destDir);
		else
			CopyFile(entry, destDir);
		}
	else if (entry.IsSymLink())
		{
		// Copy/Move link.
		if (move)
			MoveLink(entry, destDir);
		else
			CopyLink(entry, destDir);
		}
	else
		{
		MessageBox("Found an unknown.");
		}
	}

int CountDirectory(BEntry &entry)
	{
	BEntry tmpEntry;
	status_t status;
	BDirectory testDir;
	BDirectory sourceDir(&entry);
	int size = 1;	// Count directory itself.

	sourceDir.Rewind();
	status = sourceDir.GetNextEntry(&tmpEntry);
	while (status != B_ENTRY_NOT_FOUND)
		{
		if (tmpEntry.IsDirectory())
			{
			size += CountDirectory(tmpEntry);
			}
		else if (tmpEntry.IsFile())
			{
			size += CountFile(tmpEntry);
			}
		else if (tmpEntry.IsSymLink())
			{
			size += 1;
			}
		status = sourceDir.GetNextEntry(&tmpEntry);
		}
	return size;
	}

int CountFile(BEntry &entry)
	{
	BFile file(&entry, B_READ_ONLY);
	off_t size;

	file.GetSize(&size);
	return (size / BUFFER_SIZE) + 1;
	}

int CountEntry(BEntry &entry)
	{
	if (entry.IsDirectory())
		{
		return CountDirectory(entry);
		}
	else if (entry.IsFile())
		{
		return CountFile(entry);
		}
	else if (entry.IsSymLink())
		{
		return 1;
		}
	else
		return 0;
	}

void HandleRefs(entry_ref &dir_ref, BMessage *sourceMsg, bool move)
	{
	int refs;
	entry_ref file_ref;
	BEntry entry(&dir_ref);
	BDirectory destDir(&dir_ref);
	uint8 *buffer = NULL;
	int size;

	buffer = new uint8[BUFFER_SIZE];

	// Count 'em up for status window.
	size = 0;
	for (refs=0; sourceMsg->FindRef("refs", refs, &file_ref) == B_NO_ERROR; refs++)
		{
		entry.SetTo(&file_ref);
		size += CountEntry(entry);
		}

	// Create and setup the Status Window.
	statusWindow = new BWindow(BRect(50, 50, 300, 100), "xcv Progress",
			B_TITLED_WINDOW_LOOK, B_FLOATING_APP_WINDOW_FEEL,
			B_NOT_ZOOMABLE | B_NOT_RESIZABLE | B_NOT_CLOSABLE,
			B_CURRENT_WORKSPACE);
	statusBar = new BStatusBar(BRect(10, 10, 240, 60), "files");
	statusBar->SetMaxValue(size);
	statusWindow->AddChild(statusBar);
	statusWindow->Show();

	for (refs=0; sourceMsg->FindRef("refs", refs, &file_ref) == B_NO_ERROR; refs++)
		{
		entry.SetTo(&file_ref);
		DoEntry(entry, destDir, move);
		}

	statusWindow->Lock();
	statusWindow->Quit();
	delete[] buffer;
	}

#endif // defined(APP_PASTE)

void MessageBox(const char *message)
	{
	BAlert *alert = new BAlert(APP_NAME, message, "OK");
	alert->Go();
	}

void process_refs(entry_ref dir_ref, BMessage* msg, void*)
	{
	BMessage *clipMsg = NULL;
	BMessage refsMsg;

	// Try out the clipboard
	if (be_clipboard->Lock())
		{
		// Clipboard ready.
		//MessageBox("Clipboard locked.");

		clipMsg = be_clipboard->Data();
		
		#if defined(APP_CUT)
		be_clipboard->Clear();
		clipMsg->AddMessage(MSG_CUT_NAME, msg);
		#endif

		#if defined(APP_COPY)
		be_clipboard->Clear();
		clipMsg->AddMessage(MSG_COPY_NAME, msg);
		#endif

		#if defined(APP_PASTE)
		if (!clipMsg->FindMessage(MSG_CUT_NAME, &refsMsg))
			{
			//MessageBox("Found cut files");
			HandleRefs(dir_ref, &refsMsg, true);
			msg = &refsMsg;
			be_clipboard->Clear();
			}
		else if (!clipMsg->FindMessage(MSG_COPY_NAME, &refsMsg))
			{
			//MessageBox("Found copied files");
			HandleRefs(dir_ref, &refsMsg, false);
			msg = &refsMsg;
			be_clipboard->Clear();
			}
		#endif

		be_clipboard->Commit();
		be_clipboard->Unlock();
		//MessageBox("Clipboard unlocked.");
		}
	}

int main()
	{
	new BApplication(APP_SIG);
	MessageBox("This is a Tracker add-on. Move this file to:\n"
		"/boot/home/config/add-ons/Tracker");
	delete be_app;

	return 0;
	}
